////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                                                            //
//                            DiagHam  version 0.01                           //
//                                                                            //
//                  Copyright (C) 2001-2002 Nicolas Regnault                  //
//                                                                            //
//                                                                            //
//                        class of block diagonal matrix                      //
//                                                                            //
//                        last modification : 16/05/2001                      //
//                                                                            //
//                                                                            //
//    This program is free software; you can redistribute it and/or modify    //
//    it under the terms of the GNU General Public License as published by    //
//    the Free Software Foundation; either version 2 of the License, or       //
//    (at your option) any later version.                                     //
//                                                                            //
//    This program is distributed in the hope that it will be useful,         //
//    but WITHOUT ANY WARRANTY; without even the implied warranty of          //
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           //
//    GNU General Public License for more details.                            //
//                                                                            //
//    You should have received a copy of the GNU General Public License       //
//    along with this program; if not, write to the Free Software             //
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////


#ifndef BLOCKDIAGONALMATRIX_H
#define BLOCKDIAGONALMATRIX_H


#include "config.h"
#include "Matrix/Matrix.h"
#include "GeneralTools/List.h"
#include "GeneralTools/GarbageFlag.h"
#include "Output/MathematicaOutput.h"

#include <iostream>


using std::ostream;


class RealVector;
class ComplexVector;
class RealSymmetricMatrix;
class RealAntisymmetricMatrix;
class TensorProductRealVector;


class BlockDiagonalMatrix : public Matrix
{

  friend class RealVector;
  friend class ComplexVector;
  friend class RealSymmetricMatrix;
  friend class RealAntisymmetricMatrix;
  friend class TensorProductRealVector;

 private:

  double Dummy;
  int* BlockRowPosition;
  int* BlockColumnPosition;
  GarbageFlag Flag;

 protected:

  List<Matrix*> Blocks;

 public:

  // default constructor
  //
  BlockDiagonalMatrix();

  // constructor for an empty matrix
  //
  // dimension = matrix dimension
  // zero = true if matrix has to be filled with zeros
  BlockDiagonalMatrix(int dimension, bool zero = false);

  // constructor from matrix elements (without duplicating datas)
  //
  // blocks =  reference on list of pointers to matirx to use as block
  BlockDiagonalMatrix(List<Matrix*>& blocks);

  // constructor from matrix elements and block positions (without duplicating datas)
  //
  // blocks = reference on list of pointers to matirx to use as block
  // rowPosition = array containing block row positions
  // columnPosition = array containing block column positions
  BlockDiagonalMatrix(List<Matrix*>& blocks, int* rowPosition, int* columnPosition) ;

  // copy constructor (without duplicating datas)
  //
  // M = matrix to copy
  BlockDiagonalMatrix(const BlockDiagonalMatrix& M);

  // destructor
  //
  ~BlockDiagonalMatrix();

  // assignement (without duplicating datas)
  //
  // M = matrix to copy
  // return value = reference on modified matrix
  BlockDiagonalMatrix& operator = (const BlockDiagonalMatrix& M);

  // return pointer on a clone matrix (without duplicating datas)
  //
  // retrun value = pointer on new matrix 
  Matrix* Clone ();

  // set a matrix element
  //
  // i = line position
  // j = column position
  // x = new value for matrix element
  void SetMatrixElement(int i, int j, double x);

  // set a matrix element
  //
  // i = line position
  // j = column position
  // x = new value for matrix element
  void SetMatrixElement(int i, int j, const Complex& x);

  // add a value to a matrix element
  //
  // i = line position
  // j = column position
  // x = value to add to matrix element
  void AddToMatrixElement(int i, int j, double x);

  // add a value  a matrix element
  //
  // i = line position
  // j = column position
  // x = value to add to matrix element
  void AddToMatrixElement(int i, int j, const Complex& x);

  // get reference of a given matrix element
  //
  // i = line position
  // j = column position
  // return value = reference om matrix elememt
  double& operator () (int i, int j);

  // get pointer to a given block
  //
  // i = block position
  // return value = pointer to the block
  Matrix* operator [] (int i);

  // Resize matrix
  //
  // nbrRow = new number of rows
  // nbrColumn = new number of columns
  void Resize (int nbrRow, int nbrColumn);

  // Resize matrix and set to zero all elements that have been added
  //
  // nbrRow = new number of rows
  // nbrColumn = new number of columns
  void ResizeAndClean (int nbrRow, int nbrColumn);

  // add a block
  //
  // M = pointer to the matrix to add as a block
  // return value = reference on current matrix
  BlockDiagonalMatrix& operator += (Matrix* M);

 // add two matrices
  //
  // M1 = first matrix
  // M2 = second matrix
  // return value = sum of the two matrices
  friend BlockDiagonalMatrix operator + (const BlockDiagonalMatrix& M1, 
					 const BlockDiagonalMatrix& M2);

  // substract two matrices
  //
  // M1 = first matrix
  // M2 = matrix to substract to M1
  // return value = difference of the two matrices
  friend BlockDiagonalMatrix operator - (const BlockDiagonalMatrix& M1, 
					 const BlockDiagonalMatrix& M2);

  // multiply a matrix by a real number (right multiplication)
  //
  // M = source matrix
  // x = real number to use
  // return value = product result
  friend BlockDiagonalMatrix operator * (const BlockDiagonalMatrix& M, double x);

  // multiply a matrix by a real number (left multiplication)
  //
  // M = source matrix
  // x = real number to use
  // return value = product result
  friend BlockDiagonalMatrix operator * (double x, const BlockDiagonalMatrix& M);

  // divide a matrix by a real number (right multiplication)
  //
  // M = source matrix
  // x = real number to use
  // return value = division result
  friend BlockDiagonalMatrix operator / (const BlockDiagonalMatrix& M, double x);

  // add two matrices
  //
  // M = matrix to add to current matrix
  // return value = reference on current matrix
  BlockDiagonalMatrix& operator += (const BlockDiagonalMatrix& M);

  // substract two matrices
  //
  // M = matrix to substract to current matrix
  // return value = reference on current matrix
  BlockDiagonalMatrix& operator -= (const BlockDiagonalMatrix& M);

  // multiply a matrix by a real number
  //
  // x = real number to use
  // return value = reference on current matrix
  BlockDiagonalMatrix& operator *= (double x);

  // divide a matrix by a real number
  //
  // x = real number to use
  // return value = reference on current matrix
  BlockDiagonalMatrix& operator /= (double x) ;

  // evaluate matrix element
  //
  // V1 = vector to left multiply with current matrix
  // V2 = vector to right multiply with current matrix
  // return value = corresponding matrix element
  double MatrixElement (RealVector& V1, RealVector& V2);

  // evaluate matrix trace
  //
  // return value = matrix trace 
  double Tr ();

  // evaluate matrix determinant
  //
  // return value = matrix determinant 
  double Det ();

  // Output Stream overload
  //
  // Str = reference on output stream
  // P = matrix to print
  // return value = reference on output stream
  friend ostream& operator << (ostream& Str, const BlockDiagonalMatrix& P);

  // Mathematica Output Stream overload
  //
  // Str = reference on Mathematica output stream
  // P = matrix to print
  // return value = reference on output stream
  friend MathematicaOutput& operator << (MathematicaOutput& Str, const BlockDiagonalMatrix& P);

};

#endif
